from functools import partial, wraps
from .hooks import BaseHook, Tap, HookConfig


class FunctionalTap(Tap):
    def __init__(self, name, fn, before, after):
        """
        Functional taps is used in combination with a :class:`FunctionalBaseHook`

        Args:
            name (str): Name of the tap
            fn (Callable): This will be called when the hook triggers
            before (bool): If true, this tap will be called *before* the hooked function executes
            after (bool): If true, this tap will be called *after* the hooked function executes
        """
        super(FunctionalTap, self).__init__(name, fn)
        self.before = before
        self.after = after


class FunctionalHook(BaseHook):
    """
    Functional hooks are created when :class:`CreateHook` is used to decorate a class function. When a functional
    hook is tapped, a :class:`FunctionalTap` is created. Look at :func:`FunctionalHook.call` to see how taps are
    called
    """
    HOOK_TYPE = BaseHook.FUNCTIONAL

    def tap(self, name, fn, before=True, after=True):
        """
        Creates a :class:`FunctionalTap` for this hook

        Args:
            name (str): Name of the tap
            fn (Callable): This will be called when the hook triggers
            before (bool): If true, this tap will be called *before* the hooked function executes
            after (bool): If true, this tap will be called *after* the hooked function executes
        """
        tap = FunctionalTap(
            name=name,
            fn=fn,
            before=before,
            after=after
        )
        tap = self.interceptor.register(tap) if self.interceptor else tap
        self.taps.append(tap)

    def call(self, fn_args, is_before, fn_output=None):
        """
        Triggers all taps installed on this hook.

        Taps receive predefined arguments `(fn_args, fn_output, context)`

        .. code-block:: python

           fn_args = {
             "args": *args,
             "kwargs": **kwargs
           }

           fn_output = Optional[Any]

           context = {
               'hook_type': FunctionalHook.HOOK_TYPE,
               'hook_type_label': self.label,
               'tap_name': tap.name,
               'is_before': is_before
           }

        Args:
            fn_args (dict): The arguments the hooked function was called with. `fn_args: { args: Tuple, kwargs: Dict }`
            is_before (bool): True if the hook is being called after the hooked function has executed
            fn_output (Optional[Any]): The return value of the hooked function if any. None otherwise
        """
        for tap in self.taps:
            if (tap.before and is_before) or (tap.after and not is_before):
                tap.fn(
                    fn_args=fn_args,
                    fn_output=fn_output,
                    context={
                        'hook_type': FunctionalHook.HOOK_TYPE,
                        'hook_type_label': self.label,
                        'tap_name': tap.name,
                        'is_before': is_before
                    }
                )


class HookableMixin(object):
    """
    Mixin which instantiates all the decorated class methods. This is needed for decorated class methods
    """
    def __init__(self, *args, **kwargs):
        super(HookableMixin, self).__init__(*args, **kwargs)
        self.hooks = {}
        klass = type(self)
        for method in map(partial(getattr, klass), dir(klass)):
            if hasattr(method, '_pytapable'):
                hook_config = getattr(method, '_pytapable')
                self.hooks[hook_config.name] = FunctionalHook(
                    interceptor=hook_config.interceptor
                )

    def inherit_hooks(self, hookable_instance):
        """
        Given an instance which extends the :class:`HookableMixin` class, inherits all hooks from it to expose it on
        top level

        References to the inherited hooks are added in the ``self.hooks`` dict

        Args:
            hookable_instance (HookableMixin): Instance from which to inherit hooks
        """
        self.hooks.update(hookable_instance.hooks)


class CreateHook(object):
    """
    Decorator used for creating Hooks on instance methods. It takes in a name and optionally an instance of a
    :class:`HookInterceptor`.

    .. note::
        This decorator doesn't actually create the hook. It just annotates the method. The hooks are created by the
        :class:`HookableMixin` when the class is instantiated
    """
    def __init__(self, name, interceptor=None):
        self.name = name
        self.interceptor = interceptor

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            hook = args[0].hooks[self.name]
            fn_args = {"args": args, "kwargs": kwargs}

            hook.call(fn_args=fn_args, fn_output=None, is_before=True)

            out = fn(*args, **kwargs)

            hook.call(fn_args=fn_args, fn_output=out, is_before=False)

            return out

        hook_config = HookConfig(name=self.name, interceptor=self.interceptor)
        wrapper._pytapable = hook_config
        return wrapper



from typing import Callable

from .hooks import BaseHook, Tap


class Hook(BaseHook):
    HOOK_TYPE = BaseHook.INLINE

    def tap(self, name, fn):
        """
        Creates a :class:`Tap` for this hook

        Args:
            name (basestring): name of the tapable
            fn (Callable): callable to execute when hook is triggered
        """
        tap = Tap(name=name, fn=fn)
        tap = self.interceptor.register(tap) if self.interceptor else tap
        self.taps.append(tap)

    def call(self, *args, **kwargs):
        """
        Triggers the hook which executes all the taps with arguments passed as is
        """
        for tap in self.taps:
            tap.fn(*args, **kwargs)





class HookInterceptor(object):
    """
    Interceptors allow you to intercept actions that are being performed on hooks and optionally modify it
    """

    def register(self, tap):
        """
        Triggered for each added tap and allows you to modify the tap

        Args:
            tap (Tap): The Tap that is going to be installed on the hook

        Returns:
            modified_tap (Tap): The Tap to install on the hook
        """
        return tap




import uuid
from typing import Iterable
from collections import namedtuple

HookConfig = namedtuple('HookConfig', ['name', 'interceptor'])


class Tap(object):
    """
    A Tap is an object created when you tap into a hook. It holds a reference to the function you want to
    execute when the hook triggers
    """
    def __init__(self, name, fn):
        """

        Args:
            name (str): Name of the tap
            fn (Callable): This will be called when the hook triggers
        """
        self.name = name
        self.fn = fn


class BaseHook(object):
    HOOK_TYPE = -1

    FUNCTIONAL, INLINE = range(2)  # hook types

    HOOK_TYPE_LABEL = {
        FUNCTIONAL: 'Function',
        INLINE: 'Inline'
    }

    def __init__(self, interceptor=None):
        self.taps = []
        self.interceptor = interceptor

    def tap(self, *args, **kwargs):
        pass

    def call(self, *args, **kwargs):
        pass

    def intercept(self, interceptor):
        self.interceptor = interceptor

    @property
    def label(self):
        return self.HOOK_TYPE_LABEL[self.HOOK_TYPE]


def create_hook_name(name=""):
    """
    Utility to create a unique hook name. Optionally takes in a name. The output string is the name prefixed with a
    UUID. This is useful to prevent collision in hook names when one class with hooks inherits hooks from another class

    Example:
        >>> create_hook_name()
        >>> '7087eefc-8e94-4f0a-b7d3-453062bb7a34'
        >>> create_hook_name('my_hook')
        >>> '7087eefc-8e94-4f0a-b7d3-453062bb7a34:my_hook'

    Args:
        name (Optional[str]): Name of the hook
    """
    prefix = str(uuid.uuid4())
    return "{}:{}".format(prefix, name)


def create_hook_names(*names):
    """
    Useful shortcut to create multiple unique hook names in one statement

    Example:
        >>> HOOK_MY, HOOK_UNIQUE, HOOK_HOOK = create_hook_names('my', 'unique', 'hook')
        >>> HOOK_ONE, HOOK_TWO, HOOK_THREE = create_hook_names(*range(3))

    Args:
        \*names: Argument of hook names
    Returns:
        Iterable: iterable which can be deconstructed across constants.
    """

    for name in names:
        yield create_hook_name(name)



import setuptools


with open("pytapable/version.py", "r") as f:
    __version__ = None
    exec(f.read())  # evaluate to fill in __verison__

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="PyTapable",
    version=__version__,
    author="Vidhu Bhatnagar",
    author_email="vidhu1911@gmail.com",
    description="Provides utilities to implement a hookable interface",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vidhu/pytapable",
    packages=setuptools.find_packages(exclude=['tests']),
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4',
)
